{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE FlexibleInstances #-}

module Equivalence where

import CCALib
import StreamTransformer
import Test.QuickCheck
import Control.Monad
import Prelude hiding (init)

newtype Observation a = Observation (a -> String)
instance Show (Observation a) where
  show _ = "<<Observation>>"

instance Arbitrary (Observation Double) where
  arbitrary = return $ Observation show

instance Arbitrary (Observation ()) where
  arbitrary = return $ Observation show

instance (Arbitrary (Observation a), Arbitrary (Observation b))
         => Arbitrary (Observation (a, b)) where
  arbitrary = do (Observation c) <- arbitrary
                 (Observation d) <- arbitrary
                 return $ Observation (\(a, b) -> c a ++ d b)

instance (Arbitrary a, Arbitrary (Observation b))
         => Arbitrary (Observation (a -> b)) where
  arbitrary = do a <- arbitrary
                 (Observation c) <- arbitrary
                 return $ Observation (\f -> c (f a))

instance (Arbitrary a, CoArbitrary a, Arbitrary b) => Arbitrary (SF a b) where
  arbitrary = do f <- arbitrary
                 g <- arbitrary
                 return $ loop ((second (init 0)) >>> (arr (step f g)))
    where step :: (Int -> b) -> ((a, Int) -> Int) -> (a, Int) -> (b, Int)
          step f g (a, a') = (f a', g (a, a')) 

instance (Arbitrary a, CoArbitrary b) => CoArbitrary (SF a b) where
  coarbitrary f gen = do as <- replicateM iterations arbitrary
                         coarbitrary (appSF f as) gen

instance (Arbitrary a, Arbitrary (Observation b))
         => Arbitrary (Observation (SF a b)) where
  arbitrary = do as <- replicateM iterations arbitrary
                 (Observation c) <- arbitrary
                 return $ Observation (\f -> c (appSF f as))

iterations = 5

appSF :: SF a b -> [a] -> b
appSF f as = runSF f (cycle as) !! length as
