#lang racket

(require "CCA.rkt"
         redex/reduction-semantics)

(provide make-module run-module
         haskell-expr haskell-type)

(define (make-module definitions imports exports [output-directory "."])
  (define path (make-temporary-file "Tmp~a.hs" #f output-directory))
  (define name (module-name path))
  (with-output-to-file path #:exists 'truncate
    (λ ()
      (printf "module ~a ~a where\n" name exports)
      
      (for ([i imports])
        (printf "import ~a\n" i))
      
      (for ([d definitions])
        (match-define (list x t e) d)
        (printf "~a :: ~a\n~a = ~a\n" x t x e))))
  path)

(define (run-module module-path)
  (define output (open-output-string))
  (if (parameterize ([current-input-port (open-input-string "")]
                     [current-output-port output]
                     [current-error-port output])
        (system (format "runghc ~a" module-path)))
      (get-output-string output)
      (error 'run-module (get-output-string output))))

(define (module-name module-path)
  (match (regexp-match #rx"^(.+).hs$" (file-name-from-path module-path))
    [(list _ name) name]
    [_ (error 'module-path "unable to parse module path ~a" module-path)]))

(define-metafunction CCA
  [(haskell-expr >>>) 
   (>>>)]
  [(haskell-expr (λ (x) e))
   (\\ x -> (haskell-expr e))]
  [(haskell-expr (pair e_1 e_2))
   ((haskell-expr e_1) |,| (haskell-expr e_2))]
  [(haskell-expr compose)
   (|.|)]
  [(haskell-expr :)
   (:)]
  ;; cannot use term with _s, so escape to use quote
  [(haskell-expr assoc-inv) ,'assoc_inv]
  [(haskell-expr shuffle-inv) ,'shuffle_inv]
  [(haskell-expr (any_0 ...))
   ((haskell-expr any_0) ...)]
  [(haskell-expr +) (+)]
  [(haskell-expr *) (*)]
  [(haskell-expr /) (/)]
  [(haskell-expr real) (real :: Double)]
  [(haskell-expr any) any])

(define-metafunction CCA
  [(haskell-type unit) ()]
  [(haskell-type num) Double]
  [(haskell-type (pair α β))
   ((haskell-type α) |,| (haskell-type β))]
  [(haskell-type (-> α β))
   ((haskell-type α) -> (haskell-type β))]
  [(haskell-type (arr α β))
   (SF (haskell-type α) (haskell-type β))])
