#lang scheme 
(require "lib/redex-model.rkt")
(require redex/pict)


;                                                                                                                                      
;                                                                                                                                      
;                                                                                                                                      
;                                                                                                                                      
;                   ;                                 ;;;;                                ;;;;                                         
;                   ;                                    ;                                   ;                                         
;                                                        ;                                   ;                                         
;                                                        ;                                   ;                                         
;     ;;;;; ;    ;;;;       ;; ;;  ;;   ;;  ;;;;         ;         ;;;;                      ;         ;;;;;     ;; ;;;        ;;; ;;  
;    ;     ;;       ;        ;;  ;;  ;   ; ;    ;        ;        ;    ;                     ;        ;     ;     ;;   ;      ;   ;;   
;    ;      ;       ;        ;   ;   ;   ;;      ;       ;       ;      ;                    ;              ;     ;     ;    ;     ;   
;     ;;;;;         ;        ;   ;   ;   ;       ;       ;       ;      ;    ;;;;;;;;        ;              ;     ;     ;    ;     ;   
;          ;        ;        ;   ;   ;   ;       ;       ;       ;;;;;;;;                    ;         ;;;;;;     ;     ;    ;     ;   
;           ;       ;        ;   ;   ;   ;       ;       ;       ;                           ;        ;     ;     ;     ;    ;     ;   
;    ;      ;       ;        ;   ;   ;   ;;      ;       ;       ;                           ;       ;      ;     ;     ;    ;     ;   
;    ;;    ;        ;        ;   ;   ;   ; ;    ;        ;        ;     ;                    ;       ;     ;;     ;     ;     ;   ;;   
;    ; ;;;;     ;;;;;;;;;   ;;;  ;;  ;;  ;  ;;;;     ;;;;;;;;;     ;;;;;                 ;;;;;;;;;    ;;;;; ;;   ;;;   ;;;     ;;; ;   
;                                        ;                                                                                         ;   
;                                        ;                                                                                         ;   
;                                        ;                                                                                        ;    
;                                       ;;;                                                                                   ;;;;     
;                                                                                                                                      
;                                                                                                                                      

;; the language without memoization (figure 3)

(define-language simple-lang
  (P (t e))
  (e x v (e e) (spawn e)
     mkCh (send e e) (recv e))
  (v unit (λ x e) l)
  (t number)
  (l number)
  ((x y z) variable-not-otherwise-mentioned)
  (EC hole (EC e) (v EC) (spawn EC)
      (send EC e) (send l EC) (recv EC)))


(define-metafunction simple-lang
  s-subst : x any -> any
  ((s-subst x_1 any_1 (λ x_1 any_2)) (λ x_1 any_2))
  ((s-subst x_1 any_1 (λ x_2 any_2)) (λ x_2 (s-subst x_1 any_1 any_2)))
  ((s-subst x_1 any_1 x_1) any_1)
  ((s-subst x_1 any_1 (any_2 ...)) ((s-subst x_1 any_1 any_2) ...))
  ((s-subst x_1 any_1 any_2) any_2))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                       simple-lang reduction relations
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define s-reduce-show
  (reduction-relation
   simple-lang
   (--> (P_1 P_2 ... (t (in-hole EC ((λ x e) v))))
        (P_1 P_2 ... (t (in-hole EC (subst x v e))))
        app)
   (--> (P_1 P_2 ... (t_1 (in-hole EC (spawn (λ x e)))))
        (P_1 P_2 ... (t_1 (in-hole EC unit)) (t_2 ((λ x e) unit)))
        (where t_2 (simple:fresh-t))
        spawn)
   (--> (P_1 P_2 ... (t (in-hole EC mkCh)))
        (P_1 P_2 ... (t (in-hole EC l )))
        (where l (simple:fresh-l))
        channel)
   (--> (P_1 P_2 ... (t_1 (in-hole EC_1 (send l v))) (t_2 (in-hole EC_2 (recv l))))
        (P_1 P_2 ... (t_1 (in-hole EC_1 unit)) (t_2 (in-hole EC_2 v)))
        comm)))

;; s-reduce uses facilities of library/redex.rkt to reduce all the possible thread 
;; permutations of the given state delivering a set of all the possible results 
;; of reduction on the given state. This way we mimic not-determinism with a set of all the 
;; possible results.
(define s-reduce
  (reduction-relation/*
   simple-lang
   (--> (P_1 ...)
        ,(first-possible-apply-rr 
          s-reduce-show
          (permutate (term (P_1 ...)))))))


;                                                                                                              
;                                                                                                              
;                                                                                                              
;                                                                                                              
;                                                                 ;;;;                                         
;                                                                    ;                                         
;                                                                    ;                                         
;                                                                    ;                                         
;   ;; ;;  ;;      ;;;;     ;; ;;  ;;      ;;;;;                     ;         ;;;;;     ;; ;;;        ;;; ;;  
;    ;;  ;;  ;    ;    ;     ;;  ;;  ;    ;     ;                    ;        ;     ;     ;;   ;      ;   ;;   
;    ;   ;   ;   ;      ;    ;   ;   ;   ;       ;                   ;              ;     ;     ;    ;     ;   
;    ;   ;   ;   ;      ;    ;   ;   ;   ;       ;   ;;;;;;;;        ;              ;     ;     ;    ;     ;   
;    ;   ;   ;   ;;;;;;;;    ;   ;   ;   ;       ;                   ;         ;;;;;;     ;     ;    ;     ;   
;    ;   ;   ;   ;           ;   ;   ;   ;       ;                   ;        ;     ;     ;     ;    ;     ;   
;    ;   ;   ;   ;           ;   ;   ;   ;       ;                   ;       ;      ;     ;     ;    ;     ;   
;    ;   ;   ;    ;     ;    ;   ;   ;    ;     ;                    ;       ;     ;;     ;     ;     ;   ;;   
;   ;;;  ;;  ;;    ;;;;;    ;;;  ;;  ;;    ;;;;;                 ;;;;;;;;;    ;;;;; ;;   ;;;   ;;;     ;;; ;   
;                                                                                                          ;   
;                                                                                                          ;   
;                                                                                                         ;    
;                                                                                                     ;;;;     
;                                                                                                              
;                                                                                                              

; the language with memoization (figures 4, 6 and 11)

(define-language memo-lang
  (P (θs t e) (θs t Cs e))
  (e x (λ x e) v (e e) (spawn e)
     mkCh (send e e) (recv e))
  (v unit  (λ δ x e) l)
  (t number)
  (l number)
  (δ number)
  (c (R l v) (S l v) Rt)
  (C (c e) ((Sp e) e) ((Ch l) e))
  (Cs (C ...))
  (σ (((δ v) Cs) ...))
  (θ (δ v E Cs))
  (θs (θ ...))
  ((x y z) variable-not-otherwise-mentioned)
  (E hole (E e) (v E) (spawn E)
     (send E e) (send l E) (recv E))
  (redex s-redex c-redex)
  (s-redex (λ x e) ((λ δ x e) v) mkCH (spawn n))
  (c-redex (recv v) (send v v)))

(define-metafunction memo-lang
  subst : any any any -> any
  ((subst any_1 any_2 (λ δ_1 any_1 any_3)) (λ δ_1 any_1 any_3))
  ((subst any_1 any_2 (λ δ_2 any_2 any_3)) 
   (λ δ_2 any_2 (subst any_1 any_2 any_3)))
  ((subst any_1 any_2 (λ any_1 any_3)) (λ any_1 any_3))
  ((subst any_1 any_2 (λ any_2 any_3)) (λ any_2 (subst any_1 any_2 any_3)))
  ((subst any_1 any_2 any_1) any_2)
  ((subst any_1 any_2 (any_3 ...)) ((subst any_1 any_2 any_3) ...))
  ((subst any_1 any_2 any_3) any_3))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;                       memo-lang reduction relations
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define reduce-main
  (reduction-relation
   memo-lang
   (--> ((P_1 P_2 ... (θs t (in-hole E (λ x e)))) σ)
        ((P_1 P_2 ... (θs t (in-hole E (λ δ x e)))) σ)
        (where δ (memo:fresh-δ))
        fun)
   (--> ((P_1 P_2 ... (((δ_1 v_1 E (C_1 ...)) θ ...) t v_2))
         (((δ_3 v_3) Cs_2) ...))
        ((P_1 P_2 ... ((θ ...) t (in-hole E v_2)))
         (((δ_3 v_3) Cs_2) ... ((δ_1 v_1) (C_1 ... (Rt v_2)))))
        ret)
   (--> ((P_1 P_2 ... ((θ ...) t (in-hole E ((λ δ x e) v)))) σ)
        ((P_1 P_2 ... (((δ v E ()) θ ...) t (subst x v e))) σ)
        (side-condition (term (not-in-domain (δ v)  σ))) 
        app)
   (--> ((P_1 P_2  ... (θs t_1 (in-hole E (spawn (λ δ x e))))) σ)
        ((P_1 P_2  ... 
              ((constraint-add-seq θs 
                                   ((Sp ((λ δ x e) unit))
                                    (in-hole E (spawn (λ δ x e)))))
               t_1
               (in-hole E unit))
              (() t_2 ((λ δ x e) unit)))
         σ)
        (where t_2 (memo:fresh-t)) 
        spawn)
   (--> ((P_1 P_2 ... (θs t (in-hole E mkCh))) σ)
        ((P_1 P_2 ... 
              ((constraint-add-seq θs ((Ch l) (in-hole E l))) t (in-hole E l)))
         σ)
        (where l (memo:fresh-l))
        channel)
   (--> ((P_1 P_2 ...
              (θs_1 t_1 (in-hole E_1 (send l v)))
              (θs_2 t_2 (in-hole E_2 (recv l)))) σ)
        ((P_1 P_2 ... 
              ((constraint-add-seq θs_1 
                                   ((S l v) 
                                    (in-hole E_1 (send l v))))
               t_1 
               (in-hole E_1 unit))
              ((constraint-add-seq θs_2 
                                   ((R l v)
                                    (in-hole E_2 (recv l))))
               t_2 
               (in-hole E_2 v)))
         σ)
        comm)
   ;; Constraint Matching Rules (figure 6)
   (--> ((t (((Ch l_1) e_3) C_2 ...)) 
         (P_1 P_2 ... (θs t Cs (in-hole E ((λ δ x e_2) v)))) 
         σ)
        ((t ((subst l_2 l_1 C_2) ...)) 
         (P_1 P_2 ...
              ((constraint-add-seq θs ((Ch l_1) e_3))
               t 
               Cs 
               (in-hole E ((λ δ x e_2) v))))
         σ)
        (where l_2 (memo:fresh-l))
        mch)
   (--> ((t_1 (((Sp e_4) e_3) C_2 ...)) 
         (P_1 P_2 ... (θs t_2 Cs (in-hole E ((λ δ x e_2) v)))) σ)
        ((t_1 (C_2 ...)) 
         (P_1 P_2 ...
              ((constraint-add-seq θs ((Sp e_4) e_3))
               t_1 
               Cs
               (in-hole E ((λ δ x e_2) v)))
              (() t_2  e_4))
         (where t_2 (memo:fresh-t))
         σ) 
        mspawn)
   (--> ((t_2 (((R l v) e_3) C_2 ...)) 
         (P_1 P_2 ... 
              (θs_1 t_1 (in-hole E_1 (send l v)))
              (θs_2 t_2 Cs (in-hole E_2 ((λ δ x e_2) v_2))))
         σ)
        ((t_2 (C_2 ...)) 
         (P_1 P_2 ...
              ((constraint-add-seq θs_1 ((S l v) (in-hole E_1 (send l v))))
               t_1
               (in-hole E_1 unit))
              ((constraint-add-seq θs_2 ((R l v) e_3))
               t_2
               Cs 
               (in-hole E_2 ((λ δ x e_2) v_2))))
         σ) 
        mrecv)
   (--> ((t_2 (((S l v) e_3) C_2 ...)) 
         (P_1 P_2 ... 
              (θs_2 t_2 Cs (in-hole E_2 ((λ δ x e_2) v_2)))
              (θs_1 t_1 (in-hole E_1 (recv l))))
         σ)
        ((t_2 (C_2 ...)) 
         (P_1 P_2 ...
              ((constraint-add-seq θs_2 ((S l v) e_3))
               t_2
               Cs
               (in-hole E_2 ((λ δ x e_2) v_2)))
              ((constraint-add-seq θs_1 ((R l v) (in-hole E_1 (recv l))))
               t_1 
               (in-hole E_1 v)))
         σ) 
        msend)
   ;; Scedule Aware Partial Memoization (figure 11)
   (--> ((t ()) (P_1 P_2 ...) σ)
        ((P_1 P_2 ...) σ) 
        pause-memo)
   (--> ((P_1 P_2 ... (θs t Cs (in-hole E ((λ δ x e) v)))) σ)
        ((t (s (J Cs (P_1 P_2 ...))))
         (P_1 P_2 ... 
              (θs t (!s 
                     (J Cs (P_1 P_2 ...)))
                  (in-hole E ((λ δ x e) v)))) 
         σ)
        resume-memo)
   (--> ((P_1 P_2 ... 
              (θs_1 t_1 (((S l v) e_3) C_1 ...) (in-hole E_1 
                                                         ((λ δ_1 x e_1) v_1)))
              (θs_2 t_2 (((R l v) e_4) C_2 ...) (in-hole E_2 
                                                         ((λ δ_2 x e_2) v_2))))
         σ)
        ((P_1 P_2 ...
              ((constraint-add-seq θs_1 ((S l v) e_3))
               t_1
               (C_1 ...)
               (in-hole E_1 ((λ δ_1 x e_1) v_1)))
              ((constraint-add-seq θs_2 ((R l v) e_4))
               t_2
               (C_2 ...)
               (in-hole E_2 ((λ δ_2 x e_2) v_2))))
         σ) 
        mcom))) 


;; we seperate the following two reduction rules to make easier the definition of properties
;; for random testing.

;; the memo app rule of the paper (fig. 4)
(define reduce-memo-app
  (reduction-relation
   memo-lang
   (--> ((P_1 P_2 ... (θs t (in-hole E ((λ δ x e) v)))) σ) 
        ((t (s (J (result (δ v) σ) (P_1 P_2 ...)))) 
         (P_1 P_2 ...
              (θs t (!s (J (result (δ v) σ) (P_1 P_2 ...))) 
                  (in-hole E ((λ δ x e) v))))
         σ)
        (side-condition (term (in-domain (δ v) σ)))
        memo-app)))

;; the end memo rule of the paper (fig. 4)
(define reduce-end-memo
  (reduction-relation
   memo-lang
   (--> ((t ()) (P_1 P_2 ... 
                     (θs t ((c e_1) C ...) (in-hole E ((λ δ x e_2) v)))) 
                σ)
        ((P_1 P_2 ... (θs t (in-hole E e_1))) σ) 
        end-memo)))

;; reduce-show is the union of the tree above reduction relations. 
(define reduce-show 
  (reduction-relation/given
   memo-lang
   (--> any_1 any_2 (given reduce-memo-app any_1 any_2))
   (--> any_1 any_2 (given reduce-end-memo any_1 any_2))
   (--> any_1 any_2 (given reduce-main any_1 any_2))))

;; reduce uses facilities of library/redex.rkt to reduce all the possible thread 
;; permutations of the given state delivering a set of all the possible results 
;; of reduction on the given state. This way we mimic not-determinism with a set of all the 
;; possible results.
(define reduce
  (reduction-relation/*
   memo-lang
   (--> ((P_1 ...) σ_1)
        ,(first-possible-apply-rr
          reduce-show
          (map (lambda x (append  x (list (term σ_1))))
               (permutate (term (P_1 ...))))))
   (--> ( (t Cs) (P_1 ...) σ_1)
        ,(first-possible-apply-rr
          reduce-show
          (map (lambda x 
                 (append  (list (term (t Cs)))  x (list (term σ_1))))
               (permutate (term (P_1 ...))))))))

;                                                                                                                          
;                                                                                                                          
;                                                                                                                          
;                                                                                                                          
;                                  ;;                                           ;                                          
;                                   ;                              ;            ;                                          
;                                   ;                              ;                                                       
;                                   ;                              ;                                                       
;     ;;  ;;;      ;;;;       ;;;;  ;   ;;     ;;       ;;;; ;   ;;;;;;;;    ;;;;          ;;;;;     ;; ;;;                
;      ; ;   ;    ;    ;     ;    ;;;    ;      ;     ;;    ;;     ;            ;         ;     ;     ;;   ;               
;      ;;        ;      ;   ;       ;    ;      ;    ;       ;     ;            ;        ;       ;    ;     ;              
;      ;         ;      ;   ;       ;    ;      ;    ;             ;            ;        ;       ;    ;     ;              
;      ;         ;;;;;;;;   ;       ;    ;      ;    ;             ;            ;        ;       ;    ;     ;              
;      ;         ;          ;       ;    ;      ;    ;             ;            ;        ;       ;    ;     ;              
;      ;         ;          ;       ;    ;      ;    ;       ;     ;            ;        ;       ;    ;     ;              
;      ;          ;     ;    ;    ;;;    ;     ;;     ;     ;      ;    ;;      ;         ;     ;     ;     ;              
;     ;;;;;;       ;;;;;      ;;;;  ;;    ;;;;; ;;     ;;;;;        ;;;;    ;;;;;;;;;      ;;;;;     ;;;   ;;;             
;                                                                                        
;                                                             
;                                                                                      
;                                                                                      
;                                                                                      
;                                                                                      
;    ;;                       ;;;;                                                     
;     ;                          ;                                                     
;     ;                          ;                                                     
;     ;                          ;                                                     
;     ; ;;;        ;;;;          ;      ;;  ;;;;       ;;;;       ;;  ;;;     ;;;;; ;  
;     ;;   ;      ;    ;         ;       ; ;    ;     ;    ;       ; ;   ;   ;     ;;  
;     ;     ;    ;      ;        ;       ;;      ;   ;      ;      ;;        ;      ;  
;     ;     ;    ;      ;        ;       ;       ;   ;      ;      ;          ;;;;;    
;     ;     ;    ;;;;;;;;        ;       ;       ;   ;;;;;;;;      ;               ;   
;     ;     ;    ;               ;       ;       ;   ;             ;                ;  
;     ;     ;    ;               ;       ;;      ;   ;             ;         ;      ;  
;     ;     ;     ;     ;        ;       ; ;    ;     ;     ;      ;         ;;    ;   
;    ;;;   ;;;     ;;;;;     ;;;;;;;;;   ;  ;;;;       ;;;;;      ;;;;;;     ; ;;;;    
;                                        ;                                             
;                                        ;                                             
;                                        ;                                             
;                                       ;;;                                            
;                                                                                      
;                                                                                      


;;Constraint Addition
(define-metafunction memo-lang
  constraint-add-seq : (any ...) any -> any 
  ((constraint-add-seq (θ ...) C) ((constraint-add θ C) ...) ))

(define-metafunction memo-lang
  constraint-add : any (any ...) -> any
  ((constraint-add (δ v E (C_1 ...)) C_2) (δ v E (C_1 ... C_2)))) 

;; The J function yields the set of constraints Cs which are not satisfiable
;; in the given program state. (figure 5)
;;(It also returns the set of satisfiable constraints)
(define-metafunction memo-lang
  J : any any -> (any any)
  ((J (C ...) (P_1 P_2 ...)) (J-aux (C ...) (P_1 P_2 ...) ()))) 

(define-metafunction memo-lang
  J-aux : any any any -> (any any)
  ((J-aux (((S l v) e) C_1 ...)
          (P_1 P_2 ... (θs t (in-hole E (recv l))))
          (C_2 ... ))
   (J-aux (C_1 ...) (P_1 P_2 ...) (C_2 ... ((S l v) e))))
  ((J-aux (((R l v) e) C_1 ...) 
          (P_1 P_2 ... (θs t (in-hole E (send l v)))) 
          (C_2 ... )) 
   (J-aux (C_1 ...) (P_1 P_2 ...) (C_2 ... ((R l v) e))))
  ((J-aux ((Rt v)) (P_1 P_2 ... ) (C_2 ...)) 
   (((Rt v)) (C_2 ...)))
  ((J-aux (((Ch l) e) C_1 ...) (P_1 P_2 ... ) (C_2 ...)) 
   (J-aux (C_1 ...) (P_1 P_2 ...) (C_2 ... ((Ch l) e))))
  ((J-aux (((Sp e_1) e_2) C_1 ...) (P_1 P_2 ... ) (C_2 ...)) 
   (J-aux (C_1 ...) (P_1 P_2 ...) (C_2 ...)))
  ((J-aux (C_1 ...) (P_1 P_2 ...)(C_2 ... ))
   ((C_1 ...) (C_2 ...))))

;; pair projection metafunction
(define-metafunction memo-lang
  !s : any -> any
  ((!s (any_1 any_2)) any_1))

;; pair projection metafunction
(define-metafunction memo-lang
  s : any -> any
  ((s (any_1 any_2)) any_2))

(define (permutate l)
  (foldr (lambda (first rest) 
           (foldr (lambda (x y) 
                    (append (inject first x) y)) '() rest))
         '(())
         l))

(define (inject el l)
  (cond ((empty? l) (list (list el)))
        (else (cons 
               (cons el l) 
               (map 
                (lambda (x) 
                  (cons (car l) x))
                (inject el (cdr l)))))))  


(define (first-possible-apply-rr relation term-list) 
  (foldr (lambda (first rest)
           (let ((candidate (apply-rr relation first)))
             (cond ((empty? candidate) rest)
                   (else candidate))))
         '()
         term-list))


(define (pair=? pair-1 pair-2)
  (and (= (first pair-1) (first pair-2))
       (cond ((and (number? (second pair-1)) 
                   (number? (second pair-2)))
              (= (second pair-1) (second pair-2)))
             ((and (cons? (second pair-1)) (cons? (second pair-2)))
              (= (second (second pair-1)) (second (second pair-2))))
             (else (and (symbol? (second pair-1)) 
                        (symbol? (second pair-2)))))))

;; checks if the given location does not show up in the given memo store
(define-metafunction memo-lang 
  not-in-domain : any any -> any
  ((not-in-domain any_1 any_2)  ,(not-in-domain-fun 
                                  (term any_1)
                                  (term any_2))))


(define (not-in-domain-fun pair store)
  (andmap (λ x (not (pair=? pair (first (first x))))) store))   

;; checks if the given location shows up in the given memo store
(define-metafunction memo-lang 
  in-domain : any any -> any
  ((in-domain any_1 any_2)  ,(in-domain-fun 
                              (term any_1)
                              (term any_2))))


(define (in-domain-fun pair store)
  (ormap (λ x (pair=? pair (first (first x)))) store))

(define-metafunction memo-lang 
  result : any any -> any
  ((result any_1 any_2)  ,(result-fun 
                           (term any_1)
                           (term any_2))))

(define (result-fun pair store)
  (second (first (filter 
                  (λ x (pair=? pair (first (first x)))) 
                  store))))

;; Evaluation Contexts Composition
(define-metafunction memo-lang 
  compose : (E ...) -> E
  ((compose ()) hole)
  ((compose (E_1 E_2 ...))  (in-hole (compose (E_2 ...)) E_1))) 

;;Erasure Property (figure 12)
(define-metafunction memo-lang
  T : any -> any 
  ((T ((t (C_1 ...)) 
       (P_1 P_2 ... 
            (θs t (C_2 ...) (in-hole E ((λ δ x e) v)))) σ))
   (T (P_1 P_2 ... (θs t (C_1 ... C_2 ...) (in-hole E ((λ δ x e) v))))))
  ((T (P_1 ...) )
   ((T P_1) ...))
  ((T (((δ_1 v_1 E_1 Cs_1) ...) t e))
   (t (T (in-hole (compose (E_1 ...)) e))))
  ((T (((δ_1 v_1 E_1 Cs_1) ...) t ((any e_2) Cs_2) e))
   (t (T (in-hole (compose (E_1 ...)) e_2))))
  ((T (λ δ x e)) (λ x e))
  ((T (e_1 e_2)) ((T e_1) (T e_2)))
  ((T (spawn e)) (spawn (T e)))
  ((T (send e_1 e_2)) (send (T e_1) (T e_2)))
  ((T (recv e)) (recv (T e)))
  ((T any) any))


;                                                              
;                                                              
;                                                              
;                                                              
;       ;;;;;;                                       ;;        
;      ;                                              ;        
;      ;                                              ;        
;      ;                                              ;        
;    ;;;;;;;      ;;  ;;;      ;;;;       ;;;;; ;     ; ;;;    
;      ;           ; ;   ;    ;    ;     ;     ;;     ;;   ;   
;      ;           ;;        ;      ;    ;      ;     ;     ;  
;      ;           ;         ;      ;     ;;;;;       ;     ;  
;      ;           ;         ;;;;;;;;          ;      ;     ;  
;      ;           ;         ;                  ;     ;     ;  
;      ;           ;         ;           ;      ;     ;     ;  
;      ;           ;          ;     ;    ;;    ;      ;     ;  
;    ;;;;;;;      ;;;;;;       ;;;;;     ; ;;;;      ;;;   ;;; 
;                                                              
;                                                              
;                                                              
;                                                              
;                                                              
;                                                              

;; implementation of generation of fressh names for thread ids, 
;; function memo ids and channel ids

(define init-t 10000)
(define init-l 20000)
(define init-δ 30000)


(define (restore-fresh) (begin (set! init-t 10000) 
                               (set! init-l 20000)
                               (set! init-δ 30000)))


(define (fresh-t) (begin (set! init-t (+ init-t 1)) init-t))
(define (fresh-l) (begin (set! init-l (+ init-l 1)) init-l))
(define (fresh-δ) (begin (set! init-δ (+ init-δ 1)) init-δ))


(define-metafunction memo-lang
  memo:fresh-δ : -> any
  ((memo:fresh-δ) ,(fresh-δ)))

(define-metafunction simple-lang
  simple:fresh-t : -> any
  ((simple:fresh-t) ,(fresh-t)))

(define-metafunction memo-lang
  memo:fresh-t : -> any
  ((memo:fresh-t) ,(fresh-t)))

(define-metafunction simple-lang
  simple:fresh-l : -> any
  ((simple:fresh-l) ,(fresh-l)))

(define-metafunction memo-lang
  memo:fresh-l : -> any
  ((memo:fresh-l) ,(fresh-l)))

;                                                                                      
;                                                                                      
;                                                                                      
;                                                                                      
;                                                       ;                              
;      ;                                   ;            ;                              
;      ;                                   ;                                           
;      ;                                   ;                                           
;    ;;;;;;;;      ;;;;       ;;;;; ;    ;;;;;;;;    ;;;;        ;; ;;;        ;;; ;;  
;      ;          ;    ;     ;     ;;      ;            ;         ;;   ;      ;   ;;   
;      ;         ;      ;    ;      ;      ;            ;         ;     ;    ;     ;   
;      ;         ;      ;     ;;;;;        ;            ;         ;     ;    ;     ;   
;      ;         ;;;;;;;;          ;       ;            ;         ;     ;    ;     ;   
;      ;         ;                  ;      ;            ;         ;     ;    ;     ;   
;      ;         ;           ;      ;      ;            ;         ;     ;    ;     ;   
;      ;    ;;    ;     ;    ;;    ;       ;    ;;      ;         ;     ;     ;   ;;   
;       ;;;;       ;;;;;     ; ;;;;         ;;;;    ;;;;;;;;;    ;;;   ;;;     ;;; ;   
;                                                                                  ;   
;                                                                                  ;   
;                                                                                 ;    
;                                                                             ;;;;     
;                                                                                     
;                                                                                      
;                                                                                      
;                                                                                      
;                                                                                      
;                                                                                      
;    ;;                       ;;;;                                                     
;     ;                          ;                                                     
;     ;                          ;                                                     
;     ;                          ;                                                     
;     ; ;;;        ;;;;          ;      ;;  ;;;;       ;;;;       ;;  ;;;     ;;;;; ;  
;     ;;   ;      ;    ;         ;       ; ;    ;     ;    ;       ; ;   ;   ;     ;;  
;     ;     ;    ;      ;        ;       ;;      ;   ;      ;      ;;        ;      ;  
;     ;     ;    ;      ;        ;       ;       ;   ;      ;      ;          ;;;;;    
;     ;     ;    ;;;;;;;;        ;       ;       ;   ;;;;;;;;      ;               ;   
;     ;     ;    ;               ;       ;       ;   ;             ;                ;  
;     ;     ;    ;               ;       ;;      ;   ;             ;         ;      ;  
;     ;     ;     ;     ;        ;       ; ;    ;     ;     ;      ;         ;;    ;   
;    ;;;   ;;;     ;;;;;     ;;;;;;;;;   ;  ;;;;       ;;;;;      ;;;;;;     ; ;;;;    
;                                        ;                                             
;                                        ;                                             
;                                        ;                                             
;                                       ;;;                                            
;                                                                                      
;                                                                                      


(define (get-action l-term)
  (cond ((empty? l-term) empty)
        (else (cons (first l-term) 
                    (get-action (apply-rr reduce (first l-term)))))))


(define (concat xss)
  (foldr append empty xss))

(define (concat-map f xs)
  (concat (map f xs)))

(define (apply-rr-n n rr t)
  (cond
    [(zero? n) (list t)]
    [else (concat-map (curry apply-rr rr) (apply-rr-n (sub1 n) rr t))]))


(define (pred/match red match)
  (λ (t)
    (let ((result (apply-rr* red t)))
      (cond ((empty? result) #f)
            (else (match (first result)))))))


(define (apply-more/match red ec t)
  (let ((i-result (apply-rr red t)))
    (cond
      ((empty? i-result) (list t))
      ((redex-match memo-lang
                    (term  ((P_1 ... (θs t (in-hole E v))) σ))
                    (first i-result))
       i-result)
      (else (s-apply-more/match red ec (first i-result))))))


(define (s-apply-more/match red ec t)
  (let ((i-result (apply-rr red t)))
    (cond
      ((empty? i-result) (list t))
      ((redex-match simple-lang
                    (term  ((P_1 ... (θs t (in-hole EC v))) σ))
                    (first i-result))
       i-result)
      (else (s-apply-more/match red ec (first i-result))))))


;;Threads Extractor
(define-metafunction
  memo-lang
  E : any -> any 
  ((E ((t (C_1 ...)) (P_1  ...) σ)) (P_1 ...))
  ((E ((P_1  ...) σ)) (P_1 ...)))


;;Location and Thread id Erasure for testing 
(define-metafunction
  simple-lang
  L : any -> any 
  ((L ((t_1 e_1)... )) ((0 (L e_1)) ... ))
  ((L (λ x e)) (λ x (L e)))
  ((L (e_1 e_2)) ((L e_1) (L e_2)))
  ((L (spawn e)) (spawn (L e)))
  ((L (send e_1 e_2)) (send (L e_1) (L e_2)))
  ((L (recv e)) (recv (L e)))
  ((L l) 0)
  ((L any) any))


;                                                              
;                                                              
;                                                              
;                                                              
;                                                              
;      ;                                   ;                   
;      ;                                   ;                   
;      ;                                   ;                   
;    ;;;;;;;;      ;;;;       ;;;;; ;    ;;;;;;;;     ;;;;; ;  
;      ;          ;    ;     ;     ;;      ;         ;     ;;  
;      ;         ;      ;    ;      ;      ;         ;      ;  
;      ;         ;      ;     ;;;;;        ;          ;;;;;    
;      ;         ;;;;;;;;          ;       ;               ;   
;      ;         ;                  ;      ;                ;  
;      ;         ;           ;      ;      ;         ;      ;  
;      ;    ;;    ;     ;    ;;    ;       ;    ;;   ;;    ;   
;       ;;;;       ;;;;;     ; ;;;;         ;;;;     ; ;;;;    
;                                                              
;                                                              
;                                                              
;                                                              
;                                                              
;                                                              

(define s-simple-tau-test
  (term ((10 unit)
         (20 unit)
         (30 unit))))


(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (t_3 unit)))
                   t))
                s-simple-tau-test)


(define s-simple-comm-test
  (term ((10 unit) 
         (20 (recv 2)) 
         (30 (send 2 2))))) 

(test-predicate (pred/match 
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (20 2)))
                   t))
                s-simple-comm-test)


(define s-simple-fun+app-test
  (term ((10 ((λ x unit) 4)) 
         (20 (recv 2)) 
         (30 (send 2 2)))))

(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (20 2)))
                   t))
                s-simple-fun+app-test)

(define s-simple-mkCh-test (term ((10 unit) 
                                  (20 unit) 
                                  (30 (send mkCh 2)))))


(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((10 unit) 
                     (20 unit) 
                     (30 (send l 2))))
                   t))
                s-simple-mkCh-test)


(define s-simple-spawn-test
  (term ((10 unit) 
         (20 (spawn (λ x unit))) 
         ( 30 unit))))


(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (t_3 unit)
                     (t_4 unit))
                    t))) 
                s-simple-spawn-test)


(define s-spawn-comm-test
  (term ((10 unit)
         (20 ((λ x 
                ((λ z (recv x))
                 (spawn (λ y (send x 4)))))
              mkCh)))))

(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (20 4))
                    t))) 
                s-spawn-comm-test)

;;figure 9
(define s-memo-test 
  (term ((10 unit)
         (20 (((λ f
                 (λ g
                   (((((λ one
                         (λ two
                           (λ three
                             (λ four
                               unit)))) 
                       (spawn g))
                      (f unit))
                     (spawn g))
                    (f unit))))   
               (λ x ((λ x (recv 100)) (recv 100))))
              (λ x ((λ x (send 100 1111))
                    (send 100 2222))))))))


(test-predicate (pred/match
                 s-reduce
                 (λ (t)
                   (redex-match
                    simple-lang
                    ((t_1 unit) 
                     (t_2 unit) 
                     (t_3 unit)
                     (t_4 unit))
                    t))) 
                s-memo-test)


(define simple-tau-test
  (term (((() 10 unit)
          (() 20 unit)
          (() 30 unit))
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() t_3 unit))
                     σ)
                    t))) 
                simple-tau-test)

(define simple-comm-test
  (term (( (() 10 unit) 
           (() 20 (recv 2)) 
           (() 30 (send 2 2))) 
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() 20 2))
                     σ)
                    t))) 
                simple-comm-test)

(define simple-fun+app-test
  (term (((() 10 ((λ x unit) 4)) 
          (() 20 (recv 2)) 
          (() 30 (send 2 2))) 
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() 20 2))
                     σ)
                    t))) 
                simple-fun+app-test)


(define simple-mkCh-test
  (term (((() 10 unit) 
          (() 20 unit) 
          (() 30 (send mkCh 2))) 
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() 30 (send l 2)))
                     σ)
                    t))) 
                simple-mkCh-test)


(define simple-spawn-test
  (term (((() 10 unit) 
          (() 20 (spawn (λ x unit))) 
          (() 30 unit)) 
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() t_3 unit)
                      (() t_4 unit))
                     σ)
                    t))) 
                simple-spawn-test)


(define spawn-comm-test
  (term (((() 10 unit)
          (() 20 ((λ x 
                    ((λ z (recv x))
                     (spawn (λ y (send x 4)))))
                  mkCh)))
         ())))

(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() 20 4)
                      σ)
                     t)))) 
                spawn-comm-test)

;;figure 9
(define memo-test 
  (term (((() 10 unit)
          (() 20 (((λ f
                     (λ g
                       (((((λ one
                             (λ two
                               (λ three
                                 (λ four
                                   unit)))) 
                           (spawn g))
                          (f unit))
                         (spawn g))
                        (f unit))))   
                   (λ x ((λ x (recv 100)) (recv 100))))
                  (λ x ((λ x (send 100 1111))
                        (send 100 2222))))))
         ())))


(test-predicate (pred/match
                 reduce
                 (λ (t)
                   (redex-match
                    memo-lang
                    (((() t_1 unit) 
                      (() t_2 unit) 
                      (() t_3 unit)
                      (() t_4 unit)
                      σ)
                     t)))) 
                memo-test)

(test-results)

;                                                                          
;                                                                          
;                                                                          
;                                                                          
;                                              ;;                          
;                                               ;                          
;                                               ;                          
;                                               ;                          
;     ;;  ;;;      ;;;;;     ;; ;;;       ;;;;  ;      ;;;;;    ;; ;;  ;;  
;      ; ;   ;    ;     ;     ;;   ;     ;    ;;;     ;     ;    ;;  ;;  ; 
;      ;;               ;     ;     ;   ;       ;    ;       ;   ;   ;   ; 
;      ;                ;     ;     ;   ;       ;    ;       ;   ;   ;   ; 
;      ;           ;;;;;;     ;     ;   ;       ;    ;       ;   ;   ;   ; 
;      ;          ;     ;     ;     ;   ;       ;    ;       ;   ;   ;   ; 
;      ;         ;      ;     ;     ;   ;       ;    ;       ;   ;   ;   ; 
;      ;         ;     ;;     ;     ;    ;    ;;;     ;     ;    ;   ;   ; 
;     ;;;;;;      ;;;;; ;;   ;;;   ;;;    ;;;;  ;;     ;;;;;    ;;;  ;;  ;;
;                                                                          
;                                                                          
;                                                                          
;                                                                          
;                                                                          
;                                                                                      
;                                                                                      
;                                                                                      
;                                                                                      
;                                                       ;                              
;      ;                                   ;            ;                              
;      ;                                   ;                                           
;      ;                                   ;                                           
;    ;;;;;;;;      ;;;;       ;;;;; ;    ;;;;;;;;    ;;;;        ;; ;;;        ;;; ;;  
;      ;          ;    ;     ;     ;;      ;            ;         ;;   ;      ;   ;;   
;      ;         ;      ;    ;      ;      ;            ;         ;     ;    ;     ;   
;      ;         ;      ;     ;;;;;        ;            ;         ;     ;    ;     ;   
;      ;         ;;;;;;;;          ;       ;            ;         ;     ;    ;     ;   
;      ;         ;                  ;      ;            ;         ;     ;    ;     ;   
;      ;         ;           ;      ;      ;            ;         ;     ;    ;     ;   
;      ;    ;;    ;     ;    ;;    ;       ;    ;;      ;         ;     ;     ;   ;;   
;       ;;;;       ;;;;;     ; ;;;;         ;;;;    ;;;;;;;;;    ;;;   ;;;     ;;; ;   
;                                                                                  ;   
;                                                                                  ;   
;                                                                                 ;    
;                                                                             ;;;;     
;                                                                                      
;                                                                                      


;; Unique Decomposition
(define (unique-decomposition e)
  (or (redex-match memo-lang v e)
      (redex-match memo-lang (in-hole E redex))))  

(redex-check memo-lang e (unique-decomposition (term e)))                             

;; Sequential steps reduce in any E context
(define (no-memo-seq-step s)
  (let ((match (redex-match memo-lang 
                            (( P ... (θs t (in-hole E s-redex))) σ)
                            s)))
    (or (not match) (cons? (apply-rr reduce s)))))


(redex-check memo-lang ((P ...) σ) (no-memo-seq-step (term ((P ...) σ)))
             #:attempts 10)


;; If there is potential communication, the term is reducible
(define (comm-step s)
  (let ((match
            (redex-match memo-lang 
                         (( P_1 ... 
                            (θs_1 t_1 (send l v))
                            P_2 ...
                            (θs_2 t_2 (recv l))
                            P_3 ... )
                          σ)
                         s)))
    (or (not match) (cons? (apply-rr reduce s)))))


(redex-check memo-lang 
             ((P_1 ... ) σ) 
             (comm-step (term ((P_1 ... ) σ)))
             #:attempts 10)


(define (correct t)
  (let* ((s-term (term (T ,t)))
         (result  (first (apply-rr* reduce (append (list t) (list '()))))) 
         (s-result (first (apply-rr* s-reduce s-term))))  
    (ormap (λ x (equal? (term (L (T (E ,result)))) (first x)))
           (permutate (term (L ,s-result))))))

(redex-check memo-lang 
             ( (t_1 e_1) ...) 
             (correct (term ((() t_1 e_1) ... )))
             #:attempts 10)

;; Safety theorem (section 5)
(define (safe/no-shuffle t ec)
  (let ((s-term (term (T ,t)))
        (memo-app-result 
         (apply-rr reduce-memo-app t)))
    (cond ((empty? memo-app-result) #t)
          (else
           (let ((end-memo-result
                  (apply-rr reduce-end-memo
                            (let ((rr*-a (apply-rr* 
                                          reduce-main 
                                          (first memo-app-result))))
                              (cond ((empty? rr*-a) (first memo-app-result))
                                    (else (first rr*-a)))))))
             (cond ((empty? end-memo-result) #t)
                   (else 
                    (let ((result 
                           (first  (apply-more/match reduce-main
                                                     ec
                                                     (first end-memo-result))))         
                          (s-result (first (s-apply-more/match s-reduce
                                                               ec
                                                               s-term))))
                      (equal? (term (L (T (E ,result)))) 
                              (term (L ,s-result)))))))))))


(redex-check memo-lang 
             ((P_1 ... (θs t (in-hole E ((λ δ x e) v)))) σ) 
             (safe/no-shuffle (term 
                               ((P_1 ... (θs t (in-hole E ((λ δ x e) v)))) σ))
                              (term E))
             #:attempts 10)


;                                                                                                                                      
;                                                                                                                                      
;                                                                                                                                      
;                                                                                                                                      
;                                                                                                       ;                              
;      ;                                                                       ;           ;            ;                              
;      ;                                                                       ;           ;                                           
;      ;                                                                       ;           ;                                           
;    ;;;;;;;;  ;;;;    ;;;; ;;  ;;;;       ;;;;       ;;;;; ;      ;;;;      ;;;;;;;;    ;;;;;;;;    ;;;;        ;; ;;;        ;;; ;;  
;      ;         ;      ;    ; ;    ;     ;    ;     ;     ;;     ;    ;       ;           ;            ;         ;;   ;      ;   ;;   
;      ;         ;     ;     ;;      ;   ;      ;    ;      ;    ;      ;      ;           ;            ;         ;     ;    ;     ;   
;      ;          ;    ;     ;       ;   ;      ;     ;;;;;      ;      ;      ;           ;            ;         ;     ;    ;     ;   
;      ;          ;    ;     ;       ;   ;;;;;;;;          ;     ;;;;;;;;      ;           ;            ;         ;     ;    ;     ;   
;      ;           ;  ;      ;       ;   ;                  ;    ;             ;           ;            ;         ;     ;    ;     ;   
;      ;           ;  ;      ;;      ;   ;           ;      ;    ;             ;           ;            ;         ;     ;    ;     ;   
;      ;    ;;      ;;       ; ;    ;     ;     ;    ;;    ;      ;     ;      ;    ;;     ;    ;;      ;         ;     ;     ;   ;;   
;       ;;;;        ;;       ;  ;;;;       ;;;;;     ; ;;;;        ;;;;;        ;;;;        ;;;;    ;;;;;;;;;    ;;;   ;;;     ;;; ;   
;                   ;        ;                                                                                                     ;   
;                   ;        ;                                                                                                     ;   
;                  ;         ;                                                                                                    ;    
;               ;;;;;       ;;;                                                                                               ;;;;     
;                                                                                                                                      
;                                                                                  

(define (do-typesetting)
  (render-language simple-lang "./simple-lang.ps")
  (render-reduction-relation (model-rr s-reduce-show) 
                             "./simple-lang-reduction.ps")
  
  (render-language memo-lang "./memo-lang.ps")
  (render-reduction-relation (model-rr reduce-main)
                             "./memo-lang-reduction-main.ps")
  (render-reduction-relation (model-rr reduce-memo-app) 
                             "./memo-lang-reduction-memo-app.ps")
  (render-reduction-relation (model-rr reduce-end-memo)
                             "./memo-lang-reduction-end-memo.ps")
  
  (render-metafunction J "./memo-lang-J-meta.ps")
  (render-metafunction J-aux "./memo-lang-J-aux-meta.ps")
  
  (render-metafunction constraint-add-seq "./memo-lang-add-seq-meta.ps")
  (render-metafunction constraint-add "./memo-lang-add-meta.ps")
  
  (render-metafunction T "./memo-lang-T-meta.ps")
  (render-metafunction compose "./memo-lang-compose-meta.ps"))

