#lang racket

(require redex "redex.rkt")

(provide (all-defined-out)
         (all-from-out "redex.rkt"))

(define-extended-language gadt-lang empty-lang
  
  [(x y z)   (variable-prefix x) (variable-prefix y) (variable-prefix z)]
  [(a b c)   (variable-prefix a) (variable-prefix b) (variable-prefix c)]
  [(α β γ)   (variable-prefix α) (variable-prefix β) (variable-prefix γ)]
  [(S T U)   (variable-prefix S) (variable-prefix T) (variable-prefix U)]
  [(K L M N) (variable-prefix K) (variable-prefix L) (variable-prefix M) (variable-prefix N)]
  
  [prog (program dd ... e)]
  [dd (data (T a ...) where [K (b ...) C υ ...] ...)]

  [e K
     x
     (e : τ) ;; type annotation
     (e @ φ) ;; type instantiation
     (λ x e)
     (e e)
     ;; TYPO: `g` became `x`
     (let (x = e) in e)
     ;; TYPO: `g` became `x`
     (let (x :: σ = e) in e)
     (case e of [p → e] ...)]
  [p (K x ...)]
  
  ;; context
  [κ hole (κ : τ) (κ e) (k κ) (case κ of [p → e] ...)]
  
  [k K (K : τ) (k v) ((k v) : τ)]
  [v k (λ x e) ((λ x e) : τ)]
  
  [ψ ([x := e] ...)] ;; substitution
  
  [(m n) natural]
  [∆ ([K := σ] ...)]
  [Γ ([x := σ] ...)]
  [ν α a]
  [(τ υ) ν (T τ ...) (→ τ υ)]
  [σ τ (∀ (a ...) C ⇒ τ)]
  
  [(CC DD) (τ ~ τ)]
  [(C D) (^ CC ...)]
  [FF CC (∀ [α ...] (b ...) C ⊃ F)]
  [F (^ FF ...)]
  
  [θ ([α := τ] ...) ⊥]
  [φ ([ν := τ] ...) ⊥]
  
  )

(define-metafunction gadt-lang
  & : F ... -> F
  [(& [^ FF_1 ...] ...)
   [^ FF_2 ...]
   (where {FF_2 ...} (∪ {FF_1 ...} ...))])

(define-metafunction gadt-lang
  [(domain ⊥) {}]
  [(domain {[any_domain := any_range] ...}) {any_domain ...}])

(define-metafunction gadt-lang
  [(range ⊥) {}]
  [(range {[any_domain := any_range] ...}) {any_range ...}])

(define-metafunction gadt-lang
  [(lookup {[any_1 := any_2] ... [any_3 := any_4] [any_5 := any_6] ...} any_3)
   any_4])

(define-metafunction gadt-lang
  [(build-delta (data (T a ...) where [K (b ...) C υ ...] ...))
   {[K := (∀ (a ... b ...) C ⇒ (build-arrow υ ... (T a ...)))] ...}]
  [(build-delta dd ...) (∪ (build-delta dd) ...)])

(define-metafunction gadt-lang
  [(flatten-application (e_1 e_2) e_3 ...)
   (flatten-application e_1 e_2 e_3 ...)]
  [(flatten-application e ...) (e ...)])

(define-metafunction gadt-lang
  [(flatten-arrow τ_1 ... (→ τ_2 τ_3))
   (flatten-arrow τ_1 ... τ_2 τ_3)]
  [(flatten-arrow τ ...) (→ τ ...)])

(define-metafunction gadt-lang
  [(build-arrow τ) τ]
  [(build-arrow τ υ ...) (→ τ (build-arrow υ ...))])

(define-metafunction gadt-lang
  
  ;; e
  [(subst {[x_1 := e_1] ... [x := e] [x_2 := e_2] ...} x) e]
  [(subst ψ x) x]
  [(subst ψ K) K]
  [(subst ψ (e : τ)) ((subst ψ e) : τ)]
  [(subst ψ (e @ φ)) ((subst ψ e) @ φ)]
  [(subst ψ (λ x e)) (λ x (subst (restrict {x} ψ) e))]
  [(subst ψ (e_1 e_2)) ((subst ψ e_1) (subst ψ e_2))]
  [(subst ψ (let (x = e_1) in e_2))
   (let (x = (subst (restrict {x} ψ) e_1))
     in
     (subst (restrict {x} ψ) e_2))]
  [(subst ψ (let (x :: σ = e_1) in e_2))
   (let (x :: σ = (subst (restrict {x} ψ) e_1))
     in
     (subst (restrict {x} ψ) e_2))]
  [(subst ψ (case e_0 of [(K x ...) → e] ...))
   (case (subst ψ e_0) of
     [(K x ...) → (subst (restrict {x ...} ψ) e)] ...)]
  
  )

(define-metafunction gadt-lang
  [(restrict {x_1 ... x x_2 ...} {[x_3 := e_3] ... [x := e] [x_4 := e_4] ...})
   (restrict {x_1 ... x x_2 ...} {[x_3 := e_3] ...          [x_4 := e_4] ...})]
  [(restrict {x ...} ψ) ψ])

(define-metafunction gadt-lang
  rename : φ any -> any
  
  ;; ⊥
  [(rename ⊥ any) any]
  
  ;; φ
  [(rename φ {[ν := τ] ...}) {[ν := (rename φ τ)] ...}]
  [(rename φ ⊥) ⊥]

  ;; Γ
  [(rename φ {[x := σ] ...}) {[x := (rename φ σ)] ...}]

  ;; e
  [(rename φ x) x]
  [(rename φ K) K]
  [(rename φ (e : τ)) ((rename φ e) : (rename φ τ))]
  [(rename φ_1 (e @ φ_2)) ((rename φ_1 e) @ (rename φ_1 φ_2))]
  [(rename φ (λ x e)) (λ x (rename φ e))]
  [(rename φ (e_1 e_2)) ((rename φ e_1) (rename φ e_2))]
  [(rename φ (let (x = e_1) in e_2))
   (let (x = (rename φ e_1)) in (rename φ e_2))]
  [(rename φ (let (x :: σ = e_1) in e_2))
   (let (x :: (rename φ σ) = (rename φ e_1)) in (rename φ e_2))]
  [(rename φ (case e_1 of [p → e_2] ...))
   (case (rename φ e_1) of [p → (rename φ e_2)] ...)]
  
  ;; F
  [(rename φ [^ FF ...]) [^ (rename φ FF) ...]]
  
  ;; FF
  [(rename φ (τ ~ υ)) ((rename φ τ) ~ (rename φ υ))]
  [(rename φ (∀ [α ...] (b ...) C ⊃ F))
   (∀ (∪ (fuv (rename φ α)) ...)
      (b ...)
      (rename (remove {b ...} φ) C)
      ⊃
      (rename (remove {b ...} φ) F))]
  
  ;; σ
  [(rename φ (∀ (a ...) C ⇒ τ))
   (∀ (a ...)
      (rename (remove {a ...} φ) C)
      ⇒
      (rename (remove {a ...} φ) τ))]
  [(rename φ (→ τ υ)) (→ (rename φ τ) (rename φ υ))]
  [(rename φ (T τ ...)) (T (rename φ τ) ...)]
  [(rename {[ν_1 := τ_1] ... [ν := τ] [ν_2 := τ_2] ...} ν) τ]
  [(rename φ ν) ν]
  
  )

(define-metafunction gadt-lang
  [({[any_a := any_b] ... [any_c := any_d] [any_e := any_f] ...}
    . ← .
    {[any_g := any_h] ... [any_c := any_i] [any_j := any_k] ...})
   ({[any_a := any_b] ...                  [any_e := any_f] ...}
    . ← .
    {[any_g := any_h] ... [any_c := any_i] [any_j := any_k] ...})]
  [({[any_a := any_b] ...} . ← . {[any_c := any_d] ...})
   {[any_a := any_b] ... [any_c := any_d] ...}])

(define-metafunction gadt-lang
  [(⊎ any_1 ... ⊥ any_2 ...) ⊥]
  [(⊎ any_a ...
      {[any_b := any_c] ... [any_d := any_e] [any_f := any_g] ...}
      any_h ...
      {[any_i := any_j] ... [any_d := any_e] [any_k := any_l] ...}
      any_m ...)
   (⊎ any_a ...
      {[any_b := any_c] ... [any_d := any_e] [any_f := any_g] ...}
      any_h ...
      {[any_i := any_j] ...                  [any_k := any_l] ...}
      any_m ...)]
  [(⊎ any_a ...
      {[any_b := any_c] ... [any_d := any_e] [any_f := any_g] ...}
      any_h ...
      {[any_i := any_j] ... [any_d := any_k] [any_l := any_m] ...}
      any_n ...)
   ⊥]
  [(⊎ {[any_a := any_b] ...} ...) (∪ {[any_a := any_b] ...} ...)])

(define-metafunction gadt-lang
  [(remove {ν_1 ... ν ν_2 ...} {[ν_3 := τ_3] ... [ν := τ] [ν_4 := τ_4] ...})
   (remove {ν_1 ... ν ν_2 ...} {[ν_3 := τ_3] ...          [ν_4 := τ_4] ...})]
  [(remove {ν ...} φ) φ])

(define-metafunction gadt-lang

  ;; F
  [(fv [^ FF ...]) (∪ (fv FF) ...)]
  
  ;; FF
  [(fv (∀ [α ...] (b ...) C ⊃ F)) (∖ (∪ (fv C) (fv F)) {b ...})]
  [(fv (τ ~ υ)) (∪ (fv τ) (fv υ))]
  
  ;; Γ
  [(fv ([x := σ] ...)) (∪ (fv σ) ...)]
  
  ;; σ
  [(fv (∀ (a ...) C ⇒ τ)) (∖ (∪ (fv C) (fv τ)) {a ...})]
  [(fv (→ τ υ)) (∪ (fv τ) (fv υ))]
  [(fv (T τ ...)) (∪ (fv τ) ...)]
  [(fv ν) {ν}]
  
  )

(define-metafunction gadt-lang
  [(ftv any) (tv (fv any))])

(define-metafunction gadt-lang
  tv : {ν ...} -> {a ...}
  [(tv {a ... α ν ...}) (tv {a ... ν ...})]
  [(tv {a ...}) {a ...}])

(define-metafunction gadt-lang
  [(fuv any) (uv (fv any))])

(define-metafunction gadt-lang
  uv : {ν ...} -> {α ...}
  [(uv {α ... a ν ...}) (uv {α ... ν ...})]
  [(uv {α ...}) {α ...}])

(define (test-grammar)
  (redex-test-suite
  
    (test-equal (term (domain {[a := b] [c := d]})) (term {a c}))
  
    (test-equal (term (range {[a := b] [c := d]})) (term {b d}))
  
    (test-equal (term (lookup {[a := b] [c := d]} c)) (term d))
  
    (test-equal (term (flatten-application (((x y) z) x))) (term (x y z x)))
  
    (test-equal
      (term (flatten-arrow (→ a (→ b (→ c a)))))
      (term (→ a b c a)))
  
    (test-equal (term (build-arrow a b c a)) (term (→ a (→ b (→ c a)))))
  
    (test-equal (term (subst {[x := y]} x)) (term y))
    (test-equal (term (subst {[x := y]} z)) (term z))
  
    (test-equal (term (rename {[a := b]} a)) (term b))
    (test-equal (term (rename {[a := b]} c)) (term c))
  
    (test-equal (term (fv a)) (term {a}))))
